// Copyright (c) 2015, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibuild;

/// Message property information
class OmnibuildProperty {
  /// Property name
  final String name;

  /// Property docomentation
  final String documentation;

  /// Property type name
  final String type;

  /// Indicate if the property is an enumeration.
  final bool enumeration;

  /// Indicate if the property is passed as named argument.
  final bool named;

  /// Indicate the default value of the property.
  final String defaultValue;

  /// Constructor
  OmnibuildProperty(this.name, this.documentation, this.type, this.enumeration,
      this.named, this.defaultValue);

  /// Check if the given type should be validated by _getJsonValue.
  bool pureJsonData() {
    return _pureJsonData(type);
  }

  /// Recursive method for [pureJsonData]
  ///
  /// Allowed types:
  /// - num
  /// - int
  /// - double
  /// - bool
  /// - String
  /// - List<$allowedType>
  /// - Map<String, $allowedType>
  bool _pureJsonData(String type) {
    // Note that an enumeration is passed as integer.
    if (enumeration ||
        ['num', 'int', 'double', 'bool', 'String'].contains(type)) {
      return true;
    } else {
      if (type.startsWith('List<')) {
        return _pureJsonData(type.substring(5, type.length - 1));
      } else if (type.startsWith('Map<String,')) {
        return _pureJsonData(type.substring(11, type.length - 1).trimLeft());
      } else {
        return false;
      }
    }
  }
}

/// Message information
class OmnibuildMessage {
  /// Full omnibusId
  final String omnibusId;

  /// Message name
  final String name;

  /// Message documentation
  final String documentation;

  /// Message type (event, request, response)
  final OmnibusMessageType type;

  /// Message properties
  final List<OmnibuildProperty> properties;

  /// Constructor
  OmnibuildMessage(this.omnibusId, this.name, this.documentation, this.type,
      this.properties);

  /// Check if there is a property that is not pure JSON data
  ///
  /// If all types are pure JSON data this message can be initialized without
  /// type checking by [getJsonValue] (private function in Omnibus).
  bool supportsUncheckedInit() {
    return properties.where((OmnibuildProperty p) => !p.pureJsonData()).isEmpty;
  }
}

/// Message specification information
class OmnibuildSpecFile {
  /// Major version
  final int version;

  /// Parent package
  final String package;

  /// Copyright
  String copyright;

  /// Parent library
  String library;

  /// Description
  String description;

  /// Messages
  List<OmnibuildMessage> messages = new List<OmnibuildMessage>();

  /// Construct from YAML data.
  OmnibuildSpecFile.fromYaml(String yamlString, this.version, this.package) {
    // Should pass by reference.
    _parseYaml(this, yamlString);
  }

  /// Generate code.
  String generateCode() {
    return _generateCode(this);
  }
}
