// Copyright (c) 2015, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibuild;

/// Generate code from an OmnibuildSpecFile instance.
String _generateCode(OmnibuildSpecFile spec) {
  // Read YAML file.
  var code = new StringBuffer();

  // Write copyright.
  if (spec.copyright.isNotEmpty) {
    code.write(_prefixEachLine(spec.copyright, '// '));
    code.writeln();
  }

  // Write auto-generated notice.
  code.writeln('// THIS FILE IS AUTO-GENERATED, PLEASE DO NOT MODIFY BY HAND.');
  code.writeln();

  // Write part of statement.
  if (spec.library.isNotEmpty) {
    code.writeln("part of ${spec.library};");
    code.writeln();
  }

  // Write main description.
  if (spec.description.isNotEmpty) {
    code.write(_prefixEachLine(spec.description, '/// '));
    code.writeln();
  }

  // Loop through messages.
  spec.messages.forEach((OmnibuildMessage message) {
    code.write(_writeMessage(message));
  });

  // Generate formatted code.
  //
  // Note: an exception will be raised if the user specifies illegal types or
  // default values in the specification. In this case the code will fail and
  // the exception will propagate to higher-level mechanisms.
  var formatter = new DartFormatter();

  // Return generated code.
  return formatter.format(code.toString());
}

/// Write OmnibuildMessage code.
String _writeMessage(OmnibuildMessage message) {
  var code = new StringBuffer();

  // Write omnibusId constant.
  code.writeln('/// omnibusId for ${message.name}');
  code.writeln("const String ${message.name}Id = '${message.omnibusId}';");
  code.writeln();

  // Create variables for parameter data.
  var documentation = new StringBuffer(); // Parameter documentation
  var positionalParameters = []; // Positional parameters
  var optPositionalParameters = []; // Optional positional parameters
  var namedParameters = []; // Named parameters

  // Code for passing the parameters to the key-value message map
  var mapValues = [];

  // Add main message documentation.
  if (message.documentation.isNotEmpty) {
    documentation.write(_prefixEachLine(message.documentation, '/// '));
  }

  // Write arguments documentation header if there are any properties.
  if (message.properties.isNotEmpty) {
    if (documentation.isNotEmpty) {
      // Add spacing between arguments header and message main documentation.
      documentation.writeln('///');
    }
    documentation.writeln('/// *Arguments:*');
  }

  // Read properties.
  message.properties.forEach((OmnibuildProperty property) {
    // Write parameter documentation list item.
    documentation.writeln('///'); // Blank line
    documentation.writeln('/// - **${property.name}:**'); // Parameter list item
    documentation.write(_prefixEachLine(property.documentation, '///   '));

    /// Add property to a parameter group.
    if (property.named) {
      // Add parameter to named parameters.
      namedParameters
          .add('${property.type} ${property.name}: ${property.defaultValue}');
    } else {
      // Add parameter to optional positional parameters if a default value is
      // defined.
      if (property.defaultValue.isEmpty) {
        positionalParameters.add('${property.type} ${property.name}');
      } else {
        optPositionalParameters.add(
            '${property.type} ${property.name} = ${property.defaultValue}');
      }
    }

    // Add parameter to map values.
    mapValues.add(
        "'${property.name}': ${property.enumeration ? '${property.name}.index' : property.name}");
  });

  // Generate parameters string.
  var parameters = positionalParameters.join(',');
  if (optPositionalParameters.isNotEmpty) {
    parameters = (parameters.isNotEmpty ? parameters + ',' : '') +
        '[${optPositionalParameters.join(',')}]';
  }
  if (namedParameters.isNotEmpty) {
    parameters = (parameters.isNotEmpty ? parameters + ',' : '') +
        '{${namedParameters.join(',')}}';
  }

  // Message class name
  const messageClassMap = const {
    OmnibusMessageType.event: 'OmnibusEvent',
    OmnibusMessageType.request: 'OmnibusRequest',
    OmnibusMessageType.response: 'OmnibusResponse'
  };
  var messageClass = messageClassMap[message.type];

  // Message constructor
  var constructor = message.supportsUncheckedInit()
      ? '$messageClass.unchecked'
      : messageClass;

  // Write message builder function. Note that the documentation is
  // newline-terminated (_prefixEachLine uses writeln for each line).
  code.write('''
$documentation$messageClass ${message.name}($parameters) {
  return new $constructor(${message.name}Id, {${mapValues.join(',')}});
}\n\n''');

  // Return generated code.
  return code.toString();
}

/// Split [a] into lines and prefix them with [b]. Returns all lines joined
/// together using line endings.
String _prefixEachLine(String str, String prefix) {
  var out = new StringBuffer();
  var lines = LineSplitter.split(str);
  lines.forEach((String line) {
    out.writeln('$prefix$line');
  });
  return out.toString();
}
